from hashlib import sha256
from hmac import new as hmac
from typing import Any, Callable, Union

from .types import Endpoint, RequestProtocol, StringProvider


class SlackHeaders(object):
    X_SLACK_SIGNATURE = "X-Slack-Signature"
    X_SLACK_REQUEST_TIMESTAMP = "X-Slack-Request-Timestamp"


class InvalidSignatureError(ValueError):
    pass


def compute_slack_signature(
    request: RequestProtocol,
    signing_secret: str,
    version: str,
) -> str:
    """Compute a Slack signature from the given request."""
    timestamp = request.headers.get(SlackHeaders.X_SLACK_REQUEST_TIMESTAMP)
    body = request.body()
    message = f"{version}:{timestamp}:{body}"
    signature = hmac(
        bytes(signing_secret, "latin-1"),
        msg=bytes(message, "latin-1"),
        digestmod=sha256,
    )
    return signature.hexdigest()


def SignedSlackRoute(
    signing_secret: Union[str, StringProvider],
    version: str = "v0",
) -> Callable[[Endpoint], Endpoint]:
    """
    Decorator that acts as middleware to ensure that an
    incoming HTTP request is a valid issued Slack request.

    See https://api.slack.com/authentication/verifying-requests-from-slack
    """

    def wrapper(endpoint: Endpoint) -> Endpoint:
        def middleware(request: RequestProtocol) -> Any:
            if not isinstance(signing_secret, str):
                signing_secret = signing_secret()
            expected = request.headers.get(SlackHeaders.X_SLACK_SIGNATURE)
            actual = compute_slack_signature(
                request,
                signing_secret,
                version,
            )
            if actual != expected:
                raise InvalidSignatureError()
            return endpoint(request)

        return middleware

    return wrapper



from enum import Enum
from typing import List, Optional

from pydantic import AnyHttpUrl, BaseModel


class ElementType(str, Enum):
    actions = "actions"
    button = "button"
    context = "context"
    divider = "divider"
    image = "image"
    markdown = "mrkdown"
    plain_text = "plain_text"
    section = "section"


class Style(str, Enum):
    danger = "danger"
    primary = "primary"


class Block(BaseModel):
    type: ElementType


class Blocks(BaseModel):
    blocks: List[Block]


class Element(BaseModel):
    pass


class PlainText(Block):
    type: ElementType = ElementType.plain_text
    emoji: bool = True
    text: str


class Button(Block):
    type: ElementType = ElementType.button
    style: Style = Style.primary
    text: PlainText
    value: str


class Image(Block, Element):
    type: ElementType = ElementType.image
    alt_text: Optional[str] = None
    image_url: AnyHttpUrl


class Markdown(Block, Element):
    type: ElementType = ElementType.markdown
    text: str


class Context(Block):
    type: ElementType = ElementType.context
    elements: List[Element]


class Section(Block):
    type: ElementType = ElementType.section
    text: Markdown
    accessory: Optional[Image] = None


class Actions(Block):
    type: ElementType = ElementType.actions
    elements: List[Button]
